/*
 * Reinstrumenter.java
 *
 * Created on June 22, 2001, 8:40 AM
 *
 * See LICENSE file for license conditions.
 */

package residue;

import de.fub.bytecode.classfile.*;
import de.fub.bytecode.generic.*;
import java.io.*;
import java.util.*;

import residue.runtime.Monitor;
import residue.events.*;
import residue.tables.*;

/**
 * Class to reinstrument previously instrumented files.
 *
 * @author  chowells
 * @version 
 */
public class Reinstrumenter extends AbstractInstrumenter
{
	private ConstantPoolGen cp;
	private String classname;
	
	private Table table;
	private TableCreator tablec;
	
	private Set cfgnames = new HashSet();
	
	private String filename;
	
	public static void main(String [] args) throws Exception
	{
		if (args.length < 1)
		{
			System.out.println("The first command line argument should be the name");
			System.out.println("of the gretel config file to get information from");
			
			System.exit(1);
		}
		
		Reinstrumenter r = new Reinstrumenter();
		
		r.addInstrumentationEventListener(new InstrumentationEventListener()
		{
			public void instrumentationStarted(InstrumentationEvent e)
			{
				 System.out.println("Starting reinstrumentation");
			} // instrumentationStarted
			
			public void instrumentationFinished(InstrumentationEvent e)
			{
				System.out.println("Done reinstrumenting");
			} // instrumentationFinished
			
			public void classStarted(InstrumentationEvent e)
			{
				System.out.println("Reinstrumenting " + e.getClassName());
			} // classStarted
		}); // anonymous InstrumentationEventListener
		
		r.reinstrument(args[0]);
	} // main
	
	/**
	 * Reinstruments all files currently instrumented, removing
	 * all probes corresponding to lines which had been hit.
	 */
	public void reinstrument(String config) throws IOException
	{
		Monitor.getNames(config);
		
		table = Table.getTable(Monitor.getLineTableName());
		
		// create filename (String) -> methodname (String) -> Set of LineTableEntry
		Map entries = new HashMap();
		
		Iterator it = table.iterator();
		while (it.hasNext())
		{
			TableEntry te = (TableEntry)it.next();
			
			if (!entries.containsKey(te.getClassName()))
				entries.put(te.getClassName(), new HashMap());
			
			Map tab = (Map)entries.get(te.getClassName());
			if (!tab.containsKey(te.getMethodName()))
				tab.put(te.getMethodName(), new HashSet());
			
			Set currset = (Set)tab.get(te.getMethodName());
			currset.add(te);
		} // while
		
		int size = entries.size();
		
		instrumentationStarted(null, size);
		
		// create a new LineTable to replace the old
		tablec = new TableCreator();
		
		// copy roots over
		it = table.getSourceRoots().iterator();
		while (it.hasNext()) tablec.addSourceRoot((String)it.next());
		
		it = entries.keySet().iterator();
		while (it.hasNext())
		{
			size--;
			filename = (String)it.next();
			
			classStarted(filename, size + 1);
			
			JavaClass jc = new ClassParser(filename).parse();
			cp = new ConstantPoolGen(jc.getConstantPool());
			
			classname = jc.getClassName();
			
			Map classentries = (Map)entries.get(filename);
			
			// work in place
			Method [] m = jc.getMethods();
			
			for (int j = 0; j < m.length; j++)
			{
				String methodname = m[j].getName() + m[j].getSignature();
				Set methodentries = (Set)classentries.get(methodname);
				
				if (methodentries != null)
				{
					// should be safe, because if there were no elements in
					// the set, it wouldn't have been created.
					TableEntry te = (TableEntry)methodentries.iterator().next();
					tablec.startMethod(te.getClassName(), te.getMethodName(), te.getSourceName());
					
					m[j] = instrument(m[j], methodentries);
				}
			}
			jc.setConstantPool(cp.getFinalConstantPool());
			
			jc.dump(filename);
		} // while
		
		// write out new LineTable
		tablec.writeTo(Monitor.getLineTableName());
		
		// delete existing hitTable
		new File(Monitor.getHitTableName()).delete();
		
		// update the config file
		Monitor.setSize(tablec.getHitIndex());
		
		it = cfgnames.iterator();
		while (it.hasNext())
			Monitor.writeNames((String)it.next());
		
		instrumentationFinished(null, 0);
	} // reinstrument
	
	/**
	 * This takes a Method, and the Set of LineTableEntry items
	 * that go with the method.
	 */
	public Method instrument(Method m, Set entries) throws IOException
	{
		// sanity check.
		if (m.isNative() || m.isAbstract() || m.getCode() == null) return m;
		
		MethodGen mg = new MethodGen(m, classname, cp);
		InstructionList il = mg.getInstructionList();
		
		// if necessary, put an entry into the set of configs to redo.
		String methodname = m.getName() + m.getSignature();
		if (methodname.equals("<clinit>()V") || methodname.equals("main([Ljava/lang/String;)V"))
		{
			try
			{
				cfgnames.add(filename.substring(0, filename.length() - 6));
			}
			catch (ClassCastException cce)
			{
				// Correct behavior here is hard to determine.
				// I've decided to just ignore it, and move along
				// without adding anything to the list of files to
				// instrument
			}
		} // if
		
		// create LineTableEntry -> InstructionHandle lookup for probes
		Map probes = new HashMap();
		
		// create LineTableEntry -> InstructionHandle lookup for starts
		Map starts = new HashMap();
		
		// create LineTableEntry -> InstructionHandle lookup for ends
		Map ends = new HashMap();
		
		Iterator it = entries.iterator();
		while (it.hasNext())
		{
			TableEntry te = (TableEntry)it.next();
			probes.put(te, il.findHandle(te.getProbePosition()));
			starts.put(te, il.findHandle(te.getStartPosition()));
			ends.put(te, il.findHandle(te.getEndPosition()));
		}
		
		Map refIndices = new HashMap();
		
		// loop through the entries, removing the probes
		// that were hit, and renumbering the ones that
		// weren't.
		it = entries.iterator();
		while (it.hasNext())
		{
			TableEntry te = (TableEntry)it.next();
			
			if (te.isRef())
			{
				// If it was a ref, renumber if it wasn't executed.
				// otherwise, simply don't add to the new list.
				
				if (te.wasExecuted())
				{
					// Nothing required
				}
				else
				{
					int ref = tablec.addRefEntry(te, (InstructionHandle)starts.get(te),
					                             (InstructionHandle)ends.get(te), te.getLineSet());
					
					Integer r = new Integer(ref);
					
					Set s = table.getChildrenOf(te);
					Iterator targets = s.iterator();
					while (targets.hasNext())
						refIndices.put(targets.next(), r);
				}
			}
			else
			{
				InstructionHandle curr = (InstructionHandle)probes.get(te);
				InstructionTargeter [] its = curr.getTargeters();
				
				if (te.wasExecuted())
				{
					// remove probe
					if (its != null)
						for (int j = 0; j < its.length; j++)
							its[j].updateTarget(curr, curr.getNext().getNext());
					
					try { il.delete(curr, curr.getNext()); }
					catch (TargetLostException tle)
					{
						System.out.println("Target lost unexpectedly on probe removal");
						System.out.println("Method: " + te.getMethodName());
						System.out.println("Class: " + classname);
						System.out.println("Source: " + te.getSourceName());
					}
				}
				else
				{
					// renumber the probe
					
					il.insert(curr, new PUSH(cp, tablec.getHitIndex()));
					
					tablec.addHitEntry(te, (InstructionHandle)starts.get(te),
					                   curr.getPrev(),
					                   (InstructionHandle)ends.get(te),
					                   te.getLineSet());
					
					if (its != null)
						for (int j = 0; j < its.length; j++)
							its[j].updateTarget(curr, curr.getPrev());
					
					try { il.delete(curr); }
					catch (TargetLostException tle)
					{
						System.out.println("Target lost unexpectedly on probe renumber");
						System.out.println("Method: " + te.getMethodName());
						System.out.println("Class: " + classname);
						System.out.println("Source: " + te.getSourceName());
					}
				} // else
			} // else
		} // while
		
		il.setPositions();
		tablec.endMethod(refIndices);
		
		mg.setMaxStack();
		Method result = mg.getMethod();
		il.dispose();
		
		return result;
	} // instrument
	
} // class Reinstrumenter
